#include <tizen.h>
#include <app_alarm.h>
#include "$(appName).h"
#include "defines.h"

#define EDJE_LAYOUT_MAIN_VIEW_FILE_NAME "edje/$(appName).edj"

#define APP_CONTROL_OPERATION_ALARM_RECURRING \
			"http://tizen.org/appcontrol/operation/my_recurring_alarm"
#define APP_CONTROL_OPERATION_ALARM_ONTIME \
			"http://tizen.org/appcontrol/operation/my_ontime_alarm"
#define RECURRING_ALARMS_TO_BE_INVOKED 5
#define ALARM_DELAY 2
#define ALARMS_INTERVAL 600
#define TIME_STRING_FORMAT_BUFFER_SIZE 10
#define ALARM_MESSAGE_TIMEOUT ALARM_DELAY*1.0/6.0


typedef struct _appdata {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *layout;
	int ontime_alarm_id;		/* id of the alarm set at date */
	int recurring_alarm_id;		/* id of the delayed recurring alarm */
	int recurring_alarm_count;	/* recurring alarm invocations count */
} appdata_s;


/*
 * Function returns current local time in string format: HH:MM:SS
 * Obtained char buffer must be released using free() function.
 */
char* __get_current_time(void)
{
	char *time_buff = (char *)calloc(TIME_STRING_FORMAT_BUFFER_SIZE,
									sizeof(char));
	time_t current_time;

	time(&current_time);
	strftime(time_buff,
				TIME_STRING_FORMAT_BUFFER_SIZE,
				"%H:%M:%S",
				localtime(&current_time));

	return time_buff;
}


/*
 * Function creates new recurring alarm using alarm_schedule_after_delay()
 * function. If the alarm is created successfully, then it will be fired
 * with delay of ALARM_DELAY seconds and with interval of ALARMS_INTERVAL
 * seconds.
 */
void __set_recurring_alarm(appdata_s *appdata)
{
	int ret;
	app_control_h app_control;

	/*
	 * Creating new app_control handle.
	 * If this function fails, the alarm will not be scheduled.
	 */
	ret = app_control_create(&app_control);
	if (TRACE_ERROR_IF(ret, "Function app_control_create() failed."))
		return;

	/*
	 * Bind operation to created app_control handle.
	 * Custom operation name will be used to identify the alarm invokation
	 * within app_control callback and distinguish alarm types.
	 * If this function fails, the app_control handle must be released
	 * and the alarm will not be scheduled.
	 */
	ret = app_control_set_operation(app_control,
									APP_CONTROL_OPERATION_ALARM_RECURRING);
	if (TRACE_ERROR_IF(ret, "Function app_control_set_operation() failed.")) {
		app_control_destroy(app_control);
		return;
	}

	/*
	 * Bind package name to created app_control handle.
	 * Based on provided package name, the alarm will be invoked within
	 * the context of the application referenced by provided package name.
	 * If this function fails, the app_control handle must be released
	 * and the alarm will not be scheduled.
	 */
	ret = app_control_set_app_id(app_control, PACKAGE);
	if (TRACE_ERROR_IF(ret, "Function app_control_set_app_id() failed.")) {
		app_control_destroy(app_control);
		return;
	}

	/*
	 * Set the alarm for created app_control handle.
	 * The alarm will be scheduled with ALARM_DELAY delay and ALARMS_INTERVAL
	 * period. In result, the first alarm invokation will occur ofter
	 * ALARM_DELAY seconds. Subsequent alarms invokations will occur
	 * with ALARMS_INTERVAL seconds. If ALARMS_INTERVAL == 0, then the alarm
	 * will be invoked only once after ALARM_DELAY seconds.
	 * If function succeeds, the alarm ID is returned and stored in appdata_s
	 * structure.
	 */
	ret = alarm_schedule_after_delay(app_control,
										ALARM_DELAY,
										ALARMS_INTERVAL,
										&appdata->recurring_alarm_id);
	TRACE_ERROR_IF(ret, "Function alarm_schedule_after_delay() failed.");

	/*
	 * Finally, the app_control handle is released.
	 */
	ret = app_control_destroy(app_control);
	TRACE_ERROR_IF(ret, "Function app_control_destroy() failed.");
}


/*
 * Function creates new alarm set on time, using alarm_schedule_at_date()
 * function. If the alarm is created successfully, then it will be fired
 * at specified date and time.
 */
static void __set_ontime_alarm(appdata_s *appdata)
{
	int ret;
	app_control_h app_control;

	/*
	 * Creating new app_control handle.
	 * If this function fails, the alarm will not be scheduled.
	 */
	ret = app_control_create(&app_control);
	if (TRACE_ERROR_IF(ret, "Function app_control_create() failed."))
		return;

	/*
	 * Bind operation to created app_control handle.
	 * Custom operation name will be used to identify the alarm invokation
	 * within app_control callback and distinguish alarm types.
	 * If this function fails, the app_control handle must be released
	 * and the alarm will not be scheduled.
	 */
	ret = app_control_set_operation(app_control,
									APP_CONTROL_OPERATION_ALARM_ONTIME);
	if (TRACE_ERROR_IF(ret, "Function app_control_set_operation() failed.")) {
		app_control_destroy(app_control);
		return;
	}

	/*
	 * Bind package name to created app_control handle.
	 * Based on provided package name, the alarm will be invoked within
	 * the context of the application referenced by provided package name.
	 * If this function fails, the app_control handle must be released
	 * and the alarm will not be scheduled.
	 */
	ret = app_control_set_app_id(app_control, PACKAGE);
	if (TRACE_ERROR_IF(ret, "Function app_control_set_app_id() failed.")) {
		app_control_destroy(app_control);
		return;
	}

	/*
	 * Compute the time of on-time alarm invocation.
	 */
	time_t t_alarm = time(NULL) +
						ALARM_DELAY +
						ALARMS_INTERVAL * RECURRING_ALARMS_TO_BE_INVOKED;

	/*
	 * Set the alarm for created app_control handle.
	 * The alarm will be scheduled on time t_alarm without recurrency.
	 * If function succeeds, the alarm ID is returned and stored in appdata_s
	 * structure.
	 */
	ret = alarm_schedule_at_date(app_control,
								localtime(&t_alarm),
								0,
								&appdata->ontime_alarm_id);
	TRACE_ERROR_IF(ret, "Function alarm_schedule_at_date() failed.");

	/*
	 * Finally, the app_control handle is released.
	 */
	ret = app_control_destroy(app_control);
	TRACE_ERROR_IF(ret, "Function app_control_destroy() failed.");
}


/*
 * Callback function called by elapsing timer. In result, the text message
 * and an image hides.
 */
static Eina_Bool __recurring_alarm_timer_cb(void *data)
{
	appdata_s *ad = (appdata_s *)data;

	elm_object_signal_emit(ad->layout,
							SIGNAL_ALARM_OFF,
							PART_RECURRING_ALARM_STATE_TEXT);
	elm_object_signal_emit(ad->layout,
							SIGNAL_ALARM_OFF,
							PART_ALARM_IMAGE);

	return ECORE_CALLBACK_CANCEL;
}


/*
 * Function requests for text message and related image to show on alarm fire.
 */
static void
__emit_alarm_fired_signal(appdata_s *appdata, const char *signal_source)
{
	elm_object_signal_emit(appdata->layout, SIGNAL_ALARM_ON, signal_source);
	elm_object_signal_emit(appdata->layout, SIGNAL_ALARM_ON, PART_ALARM_IMAGE);
}


/*
 * Function is called if alarm control arrives and is identified
 * as recurring alarm.
 */
static void __recurring_alarm_invoked(app_control_h app_control,
										appdata_s *appdata)
{
	int ret;
	char *alarm_data = NULL;

	/*
	 * Get data attached to the app_control handle by using common alarm data
	 * identifier APP_CONTROL_DATA_ALARM_ID. The extracted data contains alarm
	 * identifier stored as char buffer.
	 * If this function fails, the alarm can not be properly verified.
	 */
	ret = app_control_get_extra_data(app_control,
										APP_CONTROL_DATA_ALARM_ID,
										&alarm_data);
	if (TRACE_ERROR_IF(ret, "Function app_control_get_extra_data() failed."))
		return;

	if (TRACE_ERROR_IF(!alarm_data, "alarm_data == NULL"))
		return;
	/*
	 * Verify if obtained alarm identifier is equal to the identifier
	 * of created alarm. If IDs does not match, then the alarm invocation
	 * belongs to other recurring alarm.
	 */
	if (atoi(alarm_data) != appdata->recurring_alarm_id)
		return;

	/*
	 * Count number of recurring alarm invocations
	 */
	appdata->recurring_alarm_count++;

	/*
	 * Emit signal to display alarm image and text message.
	 */
	__emit_alarm_fired_signal(appdata, PART_RECURRING_ALARM_STATE_TEXT);

	/*
	 * Timer is created to hide displayed alarm image and text message
	 */
	ecore_timer_add(ALARM_MESSAGE_TIMEOUT, __recurring_alarm_timer_cb, appdata);

	char *time_str = __get_current_time();
	TRACE_INFO("Recurring alarm #%d invoked at %s",
				appdata->recurring_alarm_count, time_str);
	free(time_str);

	/*
	 * If number of recurring alarm invocations is less then expected number,
	 * then the function ends its execution and the alarm will be invoked again.
	 */
	if (appdata->recurring_alarm_count < RECURRING_ALARMS_TO_BE_INVOKED)
		return;

	/*
	 * If recurring alarm is invoked expected numer of times, then it is
	 * canceled. If this function fails, then the recurring alarm will not
	 * be stopped.
	 */
	ret = alarm_cancel(appdata->recurring_alarm_id);
	if (TRACE_ERROR_IF(ret, "Function alarm_cancel() failed."))
		return;

	TRACE_INFO("Recurring alarm canceled");
}


/*
 * Function is called if alarm control arrives and is identified as on-tim alarm.
 */
static void __ontime_alarm_invoked(app_control_h app_control, appdata_s *appdata)
{
	int ret;
	char *alarm_data = NULL;

	/*
	 * Get data attached to the app_control handle by using common alarm data
	 * identifier APP_CONTROL_DATA_ALARM_ID. The extracted data contains alarm
	 * identifier stored as char buffer.
	 * If this function fails, the alarm can not be properly verified.
	 */
	ret = app_control_get_extra_data(app_control,
										APP_CONTROL_DATA_ALARM_ID,
										&alarm_data);
	if (TRACE_ERROR_IF(ret, "Function app_control_get_extra_data() failed."))
		return;

	if (TRACE_ERROR_IF(!alarm_data, "alarm_data == NULL"))
		return;

	/*
	 * Verify if obtained alarm identifier is equal to the identifier of
	 * created alarm. If IDs does not match, then the alarm invocation belongs
	 * to other on-time alarm.
	 */
	if (atoi(alarm_data) != appdata->ontime_alarm_id)
		return;

	/*
	 * Emit signal to display alarm image and text message. Both, image
	 * and text messages remains and are never hidden.
	 */
	__emit_alarm_fired_signal(appdata, PART_ONTIME_ALARM_STATE_TEXT);

	char *time_str = __get_current_time();

	TRACE_INFO("Ontime alarm invoked at %s", time_str);
	free(time_str);

	/*
	 * If the alarm is not set as recurring, then it must not call
	 * alarm_cancel() function. In this case, the on-tim alarm is automatically
	 * released.
	 */
}


static void __win_delete_request_cb(void *data,
									Evas_Object *obj,
									void *event_info)
{
	ui_app_exit();
}


static void __win_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	appdata_s *ad = (appdata_s *)data;

	elm_win_lower(ad->win);
}


static void __get_app_resource(const char *edj_file_in,
								char *edj_path_out,
								int edj_path_max)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, edj_path_max, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}


static void __create_base_gui(appdata_s *ad)
{
	/* Window */
	ad->win = elm_win_util_standard_add(PACKAGE, PACKAGE);
	elm_win_autodel_set(ad->win, EINA_TRUE);

	if (elm_win_wm_rotation_supported_get(ad->win)) {
		int rots[4] = { 0, 90, 180, 270 };
		elm_win_wm_rotation_available_rotations_set(ad->win,
													(const int *)(&rots), 4);
	}

	evas_object_smart_callback_add(ad->win,
									"delete,request",
									__win_delete_request_cb,
									NULL);

	/* Conformant */
	ad->conform = elm_conformant_add(ad->win);
	elm_win_indicator_mode_set(ad->win, ELM_WIN_INDICATOR_SHOW);
	elm_win_indicator_opacity_set(ad->win, ELM_WIN_INDICATOR_OPAQUE);
	evas_object_size_hint_weight_set(ad->conform,
										EVAS_HINT_EXPAND,
										EVAS_HINT_EXPAND);
	elm_win_resize_object_add(ad->win, ad->conform);
	evas_object_show(ad->conform);

	/* Layout */
	char edj_path[PATH_MAX] = {0, };
	__get_app_resource(EDJE_LAYOUT_MAIN_VIEW_FILE_NAME,
						edj_path,
						(int)PATH_MAX);
	ad->layout = elm_layout_add(ad->win);
	elm_layout_file_set(ad->layout, edj_path, GROUP_MAIN);
	evas_object_size_hint_weight_set(ad->layout,
										EVAS_HINT_EXPAND,
										EVAS_HINT_EXPAND);
	eext_object_event_callback_add(ad->layout,
										EEXT_CALLBACK_BACK,
										__win_back_cb, ad);
	elm_object_content_set(ad->conform, ad->layout);

	/* Show window after base gui is set up */
	evas_object_show(ad->win);
}


static void __initialize_alarms(appdata_s *ad)
{
	/*
	 * Set recurring alarm.
	 */
	__set_recurring_alarm(ad);

	/*
	 * Set on-time alarm.
	 */
	__set_ontime_alarm(ad);
}


static bool __create_app(void *data)
{
	appdata_s *ad = (appdata_s *)data;

	__create_base_gui(ad);

	__initialize_alarms(ad);

	return true;
}


static void __control_app(app_control_h app_control, void *data)
{
	appdata_s *appdata = (appdata_s *)data;
	char *operation = NULL;
	int ret;

	/*
	 * Get operation name bind to the app_control handle.
	 * If this function fails, then it is not possible to handle app_control
	 * call.
	 */
	ret = app_control_get_operation(app_control, &operation);
	if (TRACE_ERROR_IF(ret, "Function app_control_get_operation() failed."))
		return;

	/*
	 * Based on app_control operation, then proper action is executed.
	 * In this case, only alarm related operations are supported. Below
	 * conditional statements verify if operation type match any of customly
	 * predefined alarm operation.
	 */
	if (!strncmp(APP_CONTROL_OPERATION_ALARM_RECURRING,
					operation,
					strlen(APP_CONTROL_OPERATION_ALARM_RECURRING))) {
		__recurring_alarm_invoked(app_control, appdata);
	} else if (!strncmp(APP_CONTROL_OPERATION_ALARM_ONTIME,
					operation,
					strlen(APP_CONTROL_OPERATION_ALARM_ONTIME))) {
		__ontime_alarm_invoked(app_control, appdata);
	}

	free(operation);
}


int main(int argc, char *argv[])
{
	appdata_s ad = {0,};
	ui_app_lifecycle_callback_s event_callback = {
		.create = __create_app,
		.app_control = __control_app,
	};

	int ret = ui_app_main(argc, argv, &event_callback, &ad);
	TRACE_ERROR_IF(ret, "Function ui_app_main() failed.");

	return ret;
}
